"""
    tests.unit.test_loader
    ~~~~~~~~~~~~~~~~~~~~~~

    Test Salt's loader
"""

import collections
import compileall
import copy
import inspect
import logging
import os
import shutil
import sys
import tempfile
import textwrap

import pytest

import salt.config
import salt.loader
import salt.loader.context
import salt.loader.lazy
import salt.utils.files
import salt.utils.stringutils
from tests.support.case import ModuleCase
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


def remove_bytecode(module_path):
    paths = [module_path + "c"]
    cache_tag = sys.implementation.cache_tag
    modname, ext = os.path.splitext(module_path.split(os.sep)[-1])
    paths.append(
        os.path.join(
            os.path.dirname(module_path),
            "__pycache__",
            f"{modname}.{cache_tag}.pyc",
        )
    )
    for path in paths:
        if os.path.exists(path):
            os.unlink(path)


loader_template = """
import os
from salt.utils.decorators import depends

@depends('os')
def loaded():
    return True

@depends('non_existantmodulename')
def not_loaded():
    return True
"""


class LazyLoaderTest(TestCase):
    """
    Test the loader
    """

    module_name = "lazyloadertest"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)
        cls.utils = salt.loader.utils(cls.opts)
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        # Setup the module
        self.module_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.module_dir, ignore_errors=True)
        self.module_file = os.path.join(self.module_dir, f"{self.module_name}.py")
        with salt.utils.files.fopen(self.module_file, "w") as fh:
            fh.write(salt.utils.stringutils.to_str(loader_template))
            fh.flush()
            os.fsync(fh.fileno())

        # Invoke the loader
        self.loader = salt.loader.lazy.LazyLoader(
            [self.module_dir],
            copy.deepcopy(self.opts),
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            tag="module",
        )

    def tearDown(self):
        del self.module_dir
        del self.module_file
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    @pytest.mark.slow_test
    def test_depends(self):
        """
        Test that the depends decorator works properly
        """
        # Make sure depends correctly allowed a function to load. If this
        # results in a KeyError, the decorator is broken.
        self.assertTrue(
            isinstance(
                self.loader[self.module_name + ".loaded"], salt.loader.lazy.LoadedFunc
            )
        )
        # Make sure depends correctly kept a function from loading
        self.assertTrue(self.module_name + ".not_loaded" not in self.loader)


loader_template_module = """
import my_utils

def run():
    return my_utils.run()
"""

loader_template_utils = """
def run():
    return True
"""


class LazyLoaderUtilsTest(TestCase):
    """
    Test the loader
    """

    module_name = "lazyloaderutilstest"
    utils_name = "my_utils"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        # Setup the module
        self.module_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.module_file = os.path.join(self.module_dir, f"{self.module_name}.py")
        with salt.utils.files.fopen(self.module_file, "w") as fh:
            fh.write(salt.utils.stringutils.to_str(loader_template_module))
            fh.flush()
            os.fsync(fh.fileno())

        self.utils_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.utils_file = os.path.join(self.utils_dir, f"{self.utils_name}.py")
        with salt.utils.files.fopen(self.utils_file, "w") as fh:
            fh.write(salt.utils.stringutils.to_str(loader_template_utils))
            fh.flush()
            os.fsync(fh.fileno())

    def tearDown(self):
        shutil.rmtree(self.module_dir)
        if os.path.isdir(self.module_dir):
            shutil.rmtree(self.module_dir)
        shutil.rmtree(self.utils_dir)
        if os.path.isdir(self.utils_dir):
            shutil.rmtree(self.utils_dir)
        del self.module_dir
        del self.module_file
        del self.utils_dir
        del self.utils_file

        if self.module_name in sys.modules:
            del sys.modules[self.module_name]
        if self.utils_name in sys.modules:
            del sys.modules[self.utils_name]

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def test_utils_found(self):
        """
        Test that the extra module directory is available for imports
        """
        loader = salt.loader.LazyLoader(
            [self.module_dir],
            copy.deepcopy(self.opts),
            tag="module",
            extra_module_dirs=[self.utils_dir],
        )
        self.assertTrue(
            isinstance(loader[self.module_name + ".run"], salt.loader.lazy.LoadedFunc)
        )
        self.assertTrue(loader[self.module_name + ".run"]())

    def test_utils_not_found(self):
        """
        Test that the extra module directory is not available for imports
        """
        loader = salt.loader.LazyLoader(
            [self.module_dir], copy.deepcopy(self.opts), tag="module"
        )
        self.assertTrue(self.module_name + ".run" not in loader)


class LazyLoaderVirtualEnabledTest(TestCase):
    """
    Test the base loader of salt.
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["disable_modules"] = ["pillar"]
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        self.loader = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module"),
            copy.deepcopy(self.opts),
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            tag="module",
        )

    def tearDown(self):
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    @pytest.mark.slow_test
    def test_basic(self):
        """
        Ensure that it only loads stuff when needed
        """
        # make sure it starts empty
        self.assertEqual(self.loader._dict, {})
        # get something, and make sure its a func
        self.assertTrue(inspect.isfunction(self.loader["test.ping"].func))

        # make sure we only loaded "test" functions
        for key, val in self.loader._dict.items():
            self.assertEqual(key.split(".", 1)[0], "test")

        # make sure the depends thing worked (double check of the depends testing,
        # since the loader does the calling magically
        self.assertFalse("test.missing_func" in self.loader._dict)

    def test_badkey(self):
        with self.assertRaises(KeyError):
            self.loader[None]  # pylint: disable=W0104

        with self.assertRaises(KeyError):
            self.loader[1]  # pylint: disable=W0104

    @pytest.mark.slow_test
    def test_disable(self):
        self.assertNotIn("pillar.items", self.loader)

    @pytest.mark.slow_test
    def test_len_load(self):
        """
        Since LazyLoader is a MutableMapping, if someone asks for len() we have
        to load all
        """
        self.assertEqual(self.loader._dict, {})
        len(self.loader)  # force a load all
        self.assertNotEqual(self.loader._dict, {})

    @pytest.mark.slow_test
    def test_iter_load(self):
        """
        Since LazyLoader is a MutableMapping, if someone asks to iterate we have
        to load all
        """
        self.assertEqual(self.loader._dict, {})
        # force a load all
        for key, func in self.loader.items():
            break
        self.assertNotEqual(self.loader._dict, {})

    def test_context(self):
        """
        Make sure context is shared across modules
        """
        # make sure it starts empty
        self.assertEqual(self.loader._dict, {})
        # get something, and make sure its a func
        func = self.loader["test.ping"]
        with salt.loader.context.loader_context(self.loader):
            with patch.dict(func.__globals__["__context__"], {"foo": "bar"}):
                self.assertEqual(
                    self.loader["test.echo"].__globals__["__context__"]["foo"], "bar"
                )
                self.assertEqual(
                    self.loader["grains.get"].__globals__["__context__"]["foo"], "bar"
                )

    def test_globals(self):
        with salt.loader.context.loader_context(self.loader):
            func_globals = self.loader["test.ping"].__globals__
            self.assertEqual(
                func_globals["__grains__"].value(), self.opts.get("grains", {})
            )
            self.assertEqual(
                func_globals["__pillar__"].value(), self.opts.get("pillar", {})
            )
            # the opts passed into modules is at least a subset of the whole opts
            for key, val in func_globals["__opts__"].items():
                if (
                    key in salt.config.DEFAULT_MASTER_OPTS
                    and key not in salt.config.DEFAULT_MINION_OPTS
                ):
                    # We loaded the minion opts, but somewhere in the code, the master options got pulled in
                    # Let's just not check for equality since the option won't even exist in the loaded
                    # minion options
                    continue
                if (
                    key not in salt.config.DEFAULT_MASTER_OPTS
                    and key not in salt.config.DEFAULT_MINION_OPTS
                ):
                    # This isn't even a default configuration setting, lets carry on
                    continue
                self.assertEqual(self.opts[key], val)

    def test_pack(self):
        with salt.loader.context.loader_context(self.loader):
            self.loader.pack["__foo__"] = "bar"
            func_globals = self.loader["test.ping"].__globals__
            self.assertEqual(func_globals["__foo__"].value(), "bar")

    @pytest.mark.slow_test
    def test_virtual(self):
        self.assertNotIn("test_virtual.ping", self.loader)


class LazyLoaderVirtualDisabledTest(TestCase):
    """
    Test the loader of salt without __virtual__
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        self.loader = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module"),
            copy.deepcopy(self.opts),
            tag="module",
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            virtual_enable=False,
        )

    def tearDown(self):
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.utils
        del cls.funcs
        del cls.proxy

    @pytest.mark.slow_test
    def test_virtual(self):
        self.assertTrue(
            isinstance(self.loader["test_virtual.ping"], salt.loader.lazy.LoadedFunc)
        )


class LazyLoaderWhitelistTest(TestCase):
    """
    Test the loader of salt with a whitelist
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        self.loader = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module"),
            copy.deepcopy(self.opts),
            tag="module",
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            whitelist=["test", "pillar"],
        )

    def tearDown(self):
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    @pytest.mark.slow_test
    def test_whitelist(self):
        self.assertTrue(inspect.isfunction(self.loader["test.ping"].func))
        self.assertTrue(inspect.isfunction(self.loader["pillar.get"].func))

        self.assertNotIn("grains.get", self.loader)


class LazyLoaderGrainsBlacklistTest(TestCase):
    """
    Test the loader of grains with a blacklist
    """

    def setUp(self):
        self.opts = salt.config.minion_config(None)

    def tearDown(self):
        del self.opts

    @pytest.mark.slow_test
    def test_whitelist(self):
        opts = copy.deepcopy(self.opts)
        opts["grains_blacklist"] = ["master", "os*", "ipv[46]"]

        grains = salt.loader.grains(opts)
        self.assertNotIn("master", grains)
        self.assertNotIn("os", {g[:2] for g in list(grains)})
        self.assertNotIn("ipv4", grains)
        self.assertNotIn("ipv6", grains)


class LazyLoaderSingleItem(TestCase):
    """
    Test loading a single item via the _load() function
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    def setUp(self):
        self.loader = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module"),
            copy.deepcopy(self.opts),
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            tag="module",
        )

    def tearDown(self):
        del self.loader

    def test_single_item_no_dot(self):
        """
        Checks that a KeyError is raised when the function key does not contain a '.'
        """
        key = "testing_no_dot"
        expected = f"The key '{key}' should contain a '.'"
        with self.assertRaises(KeyError) as err:
            inspect.isfunction(self.loader["testing_no_dot"])

        result = err.exception.args[0]
        assert result == expected, result


module_template = """
__load__ = ['test', 'test_alias']
__func_alias__ = dict(test_alias='working_alias')
from salt.utils.decorators import depends

def test():
    return {count}

def test_alias():
    return True

def test2():
    return True

@depends('non_existantmodulename')
def test3():
    return True

@depends('non_existantmodulename', fallback_function=test)
def test4():
    return True
"""


class LazyLoaderReloadingTest(TestCase):
    """
    Test the loader of salt with changing modules
    """

    module_name = "loadertest"
    module_key = "loadertest.test"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        # Prevent .pyc files from being written to avoid performance issues
        # when modules are reloaded quickly (see TODO comments below)
        self._original_dont_write_bytecode = sys.dont_write_bytecode
        sys.dont_write_bytecode = True

        self.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.tmp_dir, ignore_errors=True)

        self.count = 0
        opts = copy.deepcopy(self.opts)
        dirs = salt.loader._module_dirs(opts, "modules", "module")
        dirs.append(self.tmp_dir)
        self.utils = salt.loader.utils(opts)
        self.proxy = salt.loader.proxy(opts)
        self.minion_mods = salt.loader.minion_mods(opts)
        self.loader = salt.loader.LazyLoader(
            dirs,
            opts,
            tag="module",
            pack={
                "__utils__": self.utils,
                "__proxy__": self.proxy,
                "__salt__": self.minion_mods,
            },
        )

    def tearDown(self):
        # Restore original dont_write_bytecode setting
        sys.dont_write_bytecode = self._original_dont_write_bytecode

        for attrname in ("tmp_dir", "utils", "proxy", "loader", "minion_mods", "utils"):
            try:
                delattr(self, attrname)
            except AttributeError:
                continue

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def update_module(self):
        self.count += 1
        with salt.utils.files.fopen(self.module_path, "wb") as fh:
            fh.write(
                salt.utils.stringutils.to_bytes(
                    module_template.format(count=self.count)
                )
            )
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # .pyc files are prevented by sys.dont_write_bytecode = True in setUp()

    def rm_module(self):
        os.unlink(self.module_path)
        # .pyc files are prevented by sys.dont_write_bytecode = True in setUp()

    @property
    def module_path(self):
        return os.path.join(self.tmp_dir, f"{self.module_name}.py")

    @pytest.mark.slow_test
    def test_alias(self):
        """
        Make sure that you can access alias-d modules
        """
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        self.update_module()
        self.assertNotIn(f"{self.module_name}.test_alias", self.loader)
        self.assertTrue(
            isinstance(
                self.loader[f"{self.module_name}.working_alias"],
                salt.loader.lazy.LoadedFunc,
            )
        )
        self.assertTrue(
            inspect.isfunction(self.loader[f"{self.module_name}.working_alias"].func)
        )

    @pytest.mark.slow_test
    def test_clear(self):
        self.assertTrue(
            isinstance(self.loader["test.ping"], salt.loader.lazy.LoadedFunc)
        )
        self.assertTrue(inspect.isfunction(self.loader["test.ping"].func))
        self.update_module()  # write out out custom module
        self.loader.clear()  # clear the loader dict

        # force a load of our module
        self.assertTrue(
            isinstance(self.loader[self.module_key], salt.loader.lazy.LoadedFunc)
        )
        self.assertTrue(inspect.isfunction(self.loader[self.module_key].func))

        # make sure we only loaded our custom module
        # which means that we did correctly refresh the file mapping
        for k, v in self.loader._dict.items():
            self.assertTrue(k.startswith(self.module_name))

    @pytest.mark.slow_test
    def test_load(self):
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        self.update_module()
        self.assertTrue(
            isinstance(self.loader[self.module_key], salt.loader.lazy.LoadedFunc)
        )
        self.assertTrue(inspect.isfunction(self.loader[self.module_key].func))

    @pytest.mark.slow_test
    def test__load__(self):
        """
        If a module specifies __load__ we should only load/expose those modules
        """
        self.update_module()

        # ensure it doesn't exist
        self.assertNotIn(self.module_key + "2", self.loader)

    @pytest.mark.slow_test
    def test__load__and_depends(self):
        """
        If a module specifies __load__ we should only load/expose those modules
        """
        self.update_module()
        # ensure it doesn't exist
        self.assertNotIn(self.module_key + "3", self.loader)
        self.assertNotIn(self.module_key + "4", self.loader)

    @pytest.mark.slow_test
    def test_reload(self):
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        # make sure it updates correctly
        for x in range(1, 3):
            self.update_module()
            self.loader.clear()
            self.assertEqual(self.loader[self.module_key](), self.count)

        self.rm_module()
        # make sure that even if we remove the module, its still loaded until a clear
        self.assertEqual(self.loader[self.module_key](), self.count)
        self.loader.clear()
        self.assertNotIn(self.module_key, self.loader)

    def test_wrong_bytecode(self):
        """
        Checks to make sure we don't even try to load .pyc files that are for a different Python
        This should pass (the load should fail) all the time because we don't run Salt on Py 3.4 anymore
        """
        test_module_name = "test_module.cpython-34"
        filemap_save = copy.deepcopy(self.loader.file_mapping)
        self.loader.file_mapping = {
            test_module_name: (
                "/temp/path/does/not/matter/here/__pycache__/"
                + test_module_name
                + ".pyc",
                ".pyc",
                0,
            )
        }

        self.assertFalse(self.loader._load_module(test_module_name))

        self.loader.file_mapping = copy.deepcopy(filemap_save)


virtual_aliases = ("loadertest2", "loadertest3")
virtual_alias_module_template = """
__virtual_aliases__ = {}

def test():
    return True
""".format(
    virtual_aliases
)


class LazyLoaderVirtualAliasTest(TestCase):
    """
    Test the loader of salt with changing modules
    """

    module_name = "loadertest"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        opts = copy.deepcopy(self.opts)
        dirs = salt.loader._module_dirs(opts, "modules", "module")
        dirs.append(self.tmp_dir)
        self.utils = salt.loader.utils(opts)
        self.proxy = salt.loader.proxy(opts)
        self.minion_mods = salt.loader.minion_mods(opts)
        self.loader = salt.loader.LazyLoader(
            dirs,
            opts,
            tag="module",
            pack={
                "__utils__": self.utils,
                "__proxy__": self.proxy,
                "__salt__": self.minion_mods,
            },
        )

    def tearDown(self):
        del self.tmp_dir
        del self.utils
        del self.proxy
        del self.minion_mods
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def update_module(self):
        with salt.utils.files.fopen(self.module_path, "wb") as fh:
            fh.write(salt.utils.stringutils.to_bytes(virtual_alias_module_template))
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # pyc files don't like it when we change the original quickly
        # since the header bytes only contain the timestamp (granularity of seconds)
        # TODO: don't write them? Is *much* slower on re-load (~3x)
        # https://docs.python.org/2/library/sys.html#sys.dont_write_bytecode
        remove_bytecode(self.module_path)

    @property
    def module_path(self):
        return os.path.join(self.tmp_dir, f"{self.module_name}.py")

    @pytest.mark.slow_test
    def test_virtual_alias(self):
        """
        Test the __virtual_alias__ feature
        """
        self.update_module()

        mod_names = [self.module_name] + list(virtual_aliases)
        for mod_name in mod_names:
            func_name = ".".join((mod_name, "test"))
            log.debug("Running %s (dict attribute)", func_name)
            self.assertTrue(self.loader[func_name]())
            log.debug("Running %s (loader attribute)", func_name)
            self.assertTrue(getattr(self.loader, mod_name).test())


submodule_template = """
from __future__ import absolute_import

import {0}.lib

def test():
    return ({count}, {0}.lib.test())
"""

submodule_lib_template = """
def test():
    return {count}
"""


class LazyLoaderSubmodReloadingTest(TestCase):
    """
    Test the loader of salt with changing modules
    """

    module_name = "loadertestsubmod"
    module_key = "loadertestsubmod.test"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.tmp_dir, ignore_errors=True)
        os.makedirs(self.module_dir)

        self.count = 0
        self.lib_count = 0

        opts = copy.deepcopy(self.opts)
        dirs = salt.loader._module_dirs(opts, "modules", "module")
        dirs.append(self.tmp_dir)
        self.utils = salt.loader.utils(opts)
        self.proxy = salt.loader.proxy(opts)
        self.minion_mods = salt.loader.minion_mods(opts)
        self.loader = salt.loader.LazyLoader(
            dirs,
            opts,
            tag="module",
            pack={
                "__utils__": self.utils,
                "__proxy__": self.proxy,
                "__salt__": self.minion_mods,
            },
        )

    def tearDown(self):
        del self.tmp_dir
        del self.utils
        del self.proxy
        del self.minion_mods
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def update_module(self):
        self.count += 1
        with salt.utils.files.fopen(self.module_path, "wb") as fh:
            fh.write(
                salt.utils.stringutils.to_bytes(
                    submodule_template.format(self.module_name, count=self.count)
                )
            )
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # pyc files don't like it when we change the original quickly
        # since the header bytes only contain the timestamp (granularity of seconds)
        # TODO: don't write them? Is *much* slower on re-load (~3x)
        # https://docs.python.org/2/library/sys.html#sys.dont_write_bytecode
        remove_bytecode(self.module_path)

    def rm_module(self):
        os.unlink(self.module_path)
        remove_bytecode(self.module_path)

    def update_lib(self):
        self.lib_count += 1
        for modname in list(sys.modules):
            if modname.startswith(self.module_name):
                del sys.modules[modname]
        with salt.utils.files.fopen(self.lib_path, "wb") as fh:
            fh.write(
                salt.utils.stringutils.to_bytes(
                    submodule_lib_template.format(count=self.lib_count)
                )
            )
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # pyc files don't like it when we change the original quickly
        # since the header bytes only contain the timestamp (granularity of seconds)
        # TODO: don't write them? Is *much* slower on re-load (~3x)
        # https://docs.python.org/2/library/sys.html#sys.dont_write_bytecode
        remove_bytecode(self.lib_path)

    def rm_lib(self):
        for modname in list(sys.modules):
            if modname.startswith(self.module_name):
                del sys.modules[modname]
        os.unlink(self.lib_path)
        remove_bytecode(self.lib_path)

    @property
    def module_dir(self):
        return os.path.join(self.tmp_dir, self.module_name)

    @property
    def module_path(self):
        return os.path.join(self.module_dir, "__init__.py")

    @property
    def lib_path(self):
        return os.path.join(self.module_dir, "lib.py")

    @pytest.mark.slow_test
    def test_basic(self):
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        self.update_module()
        self.update_lib()
        self.loader.clear()
        self.assertIn(self.module_key, self.loader)

    @pytest.mark.slow_test
    def test_reload(self):
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        # update both the module and the lib
        for x in range(1, 3):
            self.update_lib()
            self.update_module()
            self.loader.clear()
            self.assertNotIn(self.module_key, self.loader._dict)
            self.assertIn(self.module_key, self.loader)
            self.assertEqual(
                self.loader[self.module_key](), (self.count, self.lib_count)
            )

        # update just the module
        for x in range(1, 3):
            self.update_module()
            self.loader.clear()
            self.assertNotIn(self.module_key, self.loader._dict)
            self.assertIn(self.module_key, self.loader)
            self.assertEqual(
                self.loader[self.module_key](), (self.count, self.lib_count)
            )

        # update just the lib
        for x in range(1, 3):
            self.update_lib()
            self.loader.clear()
            self.assertNotIn(self.module_key, self.loader._dict)
            self.assertIn(self.module_key, self.loader)
            self.assertEqual(
                self.loader[self.module_key](), (self.count, self.lib_count)
            )

        self.rm_module()
        # make sure that even if we remove the module, its still loaded until a clear
        self.assertEqual(self.loader[self.module_key](), (self.count, self.lib_count))
        self.loader.clear()
        self.assertNotIn(self.module_key, self.loader)

    @pytest.mark.slow_test
    def test_reload_missing_lib(self):
        # ensure it doesn't exist
        self.assertNotIn(self.module_key, self.loader)

        # update both the module and the lib
        self.update_module()
        self.update_lib()
        self.loader.clear()
        self.assertEqual(self.loader[self.module_key](), (self.count, self.lib_count))

        # remove the lib, this means we should fail to load the module next time
        self.rm_lib()
        self.loader.clear()
        self.assertNotIn(self.module_key, self.loader)


mod_template = """
def test():
    return ({val})
"""


class LazyLoaderModulePackageTest(TestCase):
    """
    Test the loader of salt with changing modules
    """

    module_name = "loadertestmodpkg"
    module_key = "loadertestmodpkg.test"

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.tmp_dir, ignore_errors=True)

        dirs = salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module")
        dirs.append(self.tmp_dir)
        self.loader = salt.loader.LazyLoader(
            dirs,
            copy.deepcopy(self.opts),
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            tag="module",
        )

    def tearDown(self):
        del self.tmp_dir
        del self.loader

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    def update_pyfile(self, pyfile, contents):
        dirname = os.path.dirname(pyfile)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with salt.utils.files.fopen(pyfile, "wb") as fh:
            fh.write(salt.utils.stringutils.to_bytes(contents))
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # pyc files don't like it when we change the original quickly
        # since the header bytes only contain the timestamp (granularity of seconds)
        # TODO: don't write them? Is *much* slower on re-load (~3x)
        # https://docs.python.org/2/library/sys.html#sys.dont_write_bytecode
        remove_bytecode(pyfile)

    def rm_pyfile(self, pyfile):
        os.unlink(pyfile)
        remove_bytecode(pyfile)

    def update_module(self, relative_path, contents):
        self.update_pyfile(os.path.join(self.tmp_dir, relative_path), contents)

    def rm_module(self, relative_path):
        self.rm_pyfile(os.path.join(self.tmp_dir, relative_path))

    @pytest.mark.slow_test
    def test_module(self):
        # ensure it doesn't exist
        self.assertNotIn("foo", self.loader)
        self.assertNotIn("foo.test", self.loader)
        self.update_module("foo.py", mod_template.format(val=1))
        self.loader.clear()
        self.assertIn("foo.test", self.loader)
        self.assertEqual(self.loader["foo.test"](), 1)

    @pytest.mark.slow_test
    def test_package(self):
        # ensure it doesn't exist
        self.assertNotIn("foo", self.loader)
        self.assertNotIn("foo.test", self.loader)
        self.update_module("foo/__init__.py", mod_template.format(val=2))
        self.loader.clear()
        self.assertIn("foo.test", self.loader)
        self.assertEqual(self.loader["foo.test"](), 2)

    @pytest.mark.slow_test
    def test_module_package_collision(self):
        # ensure it doesn't exist
        self.assertNotIn("foo", self.loader)
        self.assertNotIn("foo.test", self.loader)
        self.update_module("foo.py", mod_template.format(val=3))
        self.loader.clear()
        self.assertIn("foo.test", self.loader)
        self.assertEqual(self.loader["foo.test"](), 3)

        self.update_module("foo/__init__.py", mod_template.format(val=4))
        self.loader.clear()
        self.assertIn("foo.test", self.loader)
        self.assertEqual(self.loader["foo.test"](), 4)


deep_init_base = """
from __future__ import absolute_import
import {0}.top_lib
import {0}.top_lib.mid_lib
import {0}.top_lib.mid_lib.bot_lib

def top():
    return {0}.top_lib.test()

def mid():
    return {0}.top_lib.mid_lib.test()

def bot():
    return {0}.top_lib.mid_lib.bot_lib.test()
"""


class LazyLoaderDeepSubmodReloadingTest(TestCase):
    module_name = "loadertestsubmoddeep"
    libs = ("top_lib", "mid_lib", "bot_lib")

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.tmp_dir, ignore_errors=True)
        os.makedirs(self.module_dir)

        self.lib_count = collections.defaultdict(int)  # mapping of path -> count

        # bootstrap libs
        with salt.utils.files.fopen(
            os.path.join(self.module_dir, "__init__.py"), "w"
        ) as fh:
            # No .decode() needed here as deep_init_base is defined as str and
            # not bytes.
            fh.write(
                salt.utils.stringutils.to_str(deep_init_base.format(self.module_name))
            )
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        self.lib_paths = {}
        dir_path = self.module_dir
        for lib_name in self.libs:
            dir_path = os.path.join(dir_path, lib_name)
            self.lib_paths[lib_name] = dir_path
            os.makedirs(dir_path)
            self.update_lib(lib_name)

        opts = copy.deepcopy(self.opts)
        dirs = salt.loader._module_dirs(opts, "modules", "module")
        dirs.append(self.tmp_dir)
        self.utils = salt.loader.utils(opts)
        self.proxy = salt.loader.proxy(opts)
        self.minion_mods = salt.loader.minion_mods(opts)
        self.loader = salt.loader.LazyLoader(
            dirs,
            copy.deepcopy(opts),
            tag="module",
            pack={
                "__utils__": self.utils,
                "__proxy__": self.proxy,
                "__salt__": self.minion_mods,
            },
        )
        self.assertIn(f"{self.module_name}.top", self.loader)

    def tearDown(self):
        del self.tmp_dir
        del self.lib_paths
        del self.utils
        del self.proxy
        del self.minion_mods
        del self.loader
        del self.lib_count

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    @property
    def module_dir(self):
        return os.path.join(self.tmp_dir, self.module_name)

    def update_lib(self, lib_name):
        for modname in list(sys.modules):
            if modname.startswith(self.module_name):
                del sys.modules[modname]
        path = os.path.join(self.lib_paths[lib_name], "__init__.py")
        self.lib_count[lib_name] += 1
        with salt.utils.files.fopen(path, "wb") as fh:
            fh.write(
                salt.utils.stringutils.to_bytes(
                    submodule_lib_template.format(count=self.lib_count[lib_name])
                )
            )
            fh.flush()
            os.fsync(fh.fileno())  # flush to disk

        # pyc files don't like it when we change the original quickly
        # since the header bytes only contain the timestamp (granularity of seconds)
        # TODO: don't write them? Is *much* slower on re-load (~3x)
        # https://docs.python.org/2/library/sys.html#sys.dont_write_bytecode
        remove_bytecode(path)

    @pytest.mark.slow_test
    def test_basic(self):
        self.assertIn(f"{self.module_name}.top", self.loader)

    def _verify_libs(self):
        for lib in self.libs:
            self.assertEqual(
                self.loader[
                    "{}.{}".format(self.module_name, lib.replace("_lib", ""))
                ](),
                self.lib_count[lib],
            )

    @pytest.mark.slow_test
    def test_reload(self):
        """
        Make sure that we can reload all libraries of arbitrary depth
        """
        self._verify_libs()

        # update them all
        for lib in self.libs:
            for x in range(5):
                self.update_lib(lib)
                self.loader.clear()
                self._verify_libs()


class LoaderMultipleGlobalTest(ModuleCase):
    """
    Tests when using multiple lazyloaders
    """

    def setUp(self):
        opts = salt.config.minion_config(None)
        self.loader1 = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(opts), "modules", "module"),
            copy.deepcopy(opts),
            pack={},
            tag="module",
            loaded_base_name="salt.loader1",
        )
        self.loader2 = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(opts), "modules", "module"),
            copy.deepcopy(opts),
            pack={},
            tag="module",
            loaded_base_name="salt.loader2",
        )

    def tearDown(self):
        del self.loader1
        del self.loader2

    def test_loader_globals(self):
        """
        Test to ensure loaders do not edit
        each others loader's namespace
        """
        self.loader1.pack["__foo__"] = "bar1"
        func1 = self.loader1["test.ping"]

        self.loader2.pack["__foo__"] = "bar2"
        func2 = self.loader2["test.ping"]
        token = salt.loader.context.loader_ctxvar.set(self.loader1)
        try:
            assert func1.__globals__["__foo__"].value() == "bar1"
        finally:
            salt.loader.context.loader_ctxvar.reset(token)
        token = salt.loader.context.loader_ctxvar.set(self.loader2)
        try:
            assert func2.__globals__["__foo__"].value() == "bar2"
        finally:
            salt.loader.context.loader_ctxvar.reset(token)


class LoaderCleanupTest(ModuleCase):
    """
    Tests the loader cleanup procedures
    """

    def setUp(self):
        opts = salt.config.minion_config(None)
        self.loader1 = salt.loader.LazyLoader(
            salt.loader._module_dirs(copy.deepcopy(opts), "modules", "module"),
            copy.deepcopy(opts),
            pack={},
            tag="module",
            loaded_base_name="salt.test",
        )

    def tearDown(self):
        del self.loader1

    def test_loader_clean_modules(self):
        loaded_base_name = self.loader1.loaded_base_name
        tag = self.loader1.tag
        self.loader1.clean_modules()

        # Base stub modules are preserved as shared infrastructure
        expected_base_stubs = {
            f"{loaded_base_name}.int",
            f"{loaded_base_name}.int.{tag}",
            f"{loaded_base_name}.ext",
            f"{loaded_base_name}.ext.{tag}",
        }

        # Check that only base stubs and utils modules remain, not actual loaded modules
        for name in list(sys.modules):
            if name.startswith(loaded_base_name):
                # Base stubs are ok
                if name in expected_base_stubs:
                    continue
                # Utils modules are shared infrastructure, ok to remain
                parts = name.split(".")
                if len(parts) >= 4 and parts[3] == "utils":
                    continue
                # Actual loaded modules should be removed
                self.fail(
                    "Found a loaded module in sys.modules: {!r}. "
                    "Only base stubs and utils modules should remain. "
                    "Base stubs: {}".format(name, expected_base_stubs)
                )


class LoaderGlobalsTest(ModuleCase):
    """
    Test all of the globals that the loader is responsible for adding to modules

    This shouldn't be done here, but should rather be done per module type (in the cases where they are used)
    so they can check ALL globals that they have (or should have) access to.

    This is intended as a shorter term way of testing these so we don't break the loader
    """

    def _verify_globals(self, mod_dict):
        """
        Verify that the globals listed in the doc string (from the test) are in these modules
        """
        # find the globals
        global_vars = {}
        for val in mod_dict.values():
            # only find salty globals
            if val.__module__.startswith(salt.loader.lazy.LOADED_BASE_NAME):
                if hasattr(val, "__globals__"):
                    if hasattr(val, "__wrapped__") or "__wrapped__" in val.__globals__:
                        global_vars[val.__module__] = sys.modules[
                            val.__module__
                        ].__dict__
                    else:
                        global_vars[val.__module__] = val.__globals__

        # if we couldn't find any, then we have no modules -- so something is broken
        self.assertNotEqual(global_vars, {}, msg="No modules were loaded.")

        # get the names of the globals you should have
        func_name = inspect.stack()[1][3]
        names = next(
            iter(salt.utils.yaml.safe_load(getattr(self, func_name).__doc__).values())
        )

        # Now, test each module!
        for item in global_vars.values():
            for name in names:
                self.assertIn(name, list(item.keys()))

    def test_auth(self):
        """
        Test that auth mods have:
            - __pillar__
            - __grains__
            - __salt__
            - __context__
        """
        self._verify_globals(salt.loader.auth(self.master_opts))

    def test_runners(self):
        """
        Test that runners have:
            - __pillar__
            - __salt__
            - __opts__
            - __grains__
            - __context__
        """
        self._verify_globals(salt.loader.runner(self.master_opts))

    def test_returners(self):
        """
        Test that returners have:
            - __salt__
            - __opts__
            - __pillar__
            - __grains__
            - __context__
        """
        self._verify_globals(salt.loader.returners(self.master_opts, {}))

    def test_pillars(self):
        """
        Test that pillars have:
            - __salt__
            - __opts__
            - __pillar__
            - __grains__
            - __context__
        """
        self._verify_globals(salt.loader.pillars(self.master_opts, {}))

    def test_tops(self):
        """
        Test that tops have: []
        """
        self._verify_globals(salt.loader.tops(self.master_opts))

    def test_outputters(self):
        """
        Test that outputters have:
            - __opts__
            - __pillar__
            - __grains__
            - __context__
        """
        self._verify_globals(salt.loader.outputters(self.master_opts))

    def test_serializers(self):
        """
        Test that serializers have: []
        """
        self._verify_globals(salt.loader.serializers(self.master_opts))

    @pytest.mark.slow_test
    def test_states(self):
        """
        Test that states have:
            - __pillar__
            - __salt__
            - __opts__
            - __grains__
            - __context__
        """
        opts = salt.config.minion_config(None)
        opts["grains"] = salt.loader.grains(opts)
        utils = salt.loader.utils(opts)
        proxy = salt.loader.proxy(opts)
        funcs = salt.loader.minion_mods(opts, utils=utils, proxy=proxy)
        self._verify_globals(salt.loader.states(opts, funcs, utils, {}, proxy=proxy))

    def test_renderers(self):
        """
        Test that renderers have:
            - __salt__    # Execution functions (i.e. __salt__['test.echo']('foo'))
            - __grains__  # Grains (i.e. __grains__['os'])
            - __pillar__  # Pillar data (i.e. __pillar__['foo'])
            - __opts__    # Minion configuration options
            - __context__ # Context dict shared amongst all modules of the same type
        """
        self._verify_globals(salt.loader.render(self.master_opts, {}))


class RawModTest(TestCase):
    """
    Test the interface of raw_mod
    """

    def setUp(self):
        self.opts = salt.config.minion_config(None)

    def tearDown(self):
        del self.opts

    @pytest.mark.slow_test
    def test_basic(self):
        testmod = salt.loader.raw_mod(self.opts, "test", None)
        for k, v in testmod.items():
            self.assertEqual(k.split(".")[0], "test")

    def test_bad_name(self):
        testmod = salt.loader.raw_mod(self.opts, "module_we_do_not_have", None)
        self.assertEqual(testmod, {})


class NetworkUtilsTestCase(ModuleCase):
    def test_is_private(self):
        mod = salt.loader.raw_mod(self.minion_opts, "network", None)
        self.assertTrue(mod["network.is_private"]("10.0.0.1"), True)

    def test_is_loopback(self):
        mod = salt.loader.raw_mod(self.minion_opts, "network", None)
        self.assertTrue(mod["network.is_loopback"]("127.0.0.1"), True)


class LazyLoaderOptimizationOrderTest(TestCase):
    """
    Test the optimization order priority in the loader (PY3)
    """

    module_name = "lazyloadertest"
    module_content = textwrap.dedent(
        """\
        # -*- coding: utf-8 -*-
        from __future__ import absolute_import

        def test():
            return True
        """
    )

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.funcs
        del cls.utils
        del cls.proxy

    def setUp(self):
        # Setup the module
        self.module_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.module_dir, ignore_errors=True)
        self.module_file = os.path.join(self.module_dir, f"{self.module_name}.py")

    def tearDown(self):
        try:
            delattr(self, "loader")
        except AttributeError:
            pass

    def _get_loader(self, order=None):
        opts = copy.deepcopy(self.opts)
        if order is not None:
            opts["optimization_order"] = order
        # Return a loader
        return salt.loader.LazyLoader(
            [self.module_dir],
            opts,
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
            tag="module",
        )

    def _get_module_filename(self):
        # The act of referencing the loader entry forces the module to be
        # loaded by the LazyDict.
        mod_fullname = self.loader[next(iter(self.loader))].__module__
        return sys.modules[mod_fullname].__file__

    def _expected(self, optimize=0):
        return "lazyloadertest.cpython-{}{}{}.pyc".format(
            sys.version_info[0],
            sys.version_info[1],
            "" if not optimize else f".opt-{optimize}",
        )

    def _write_module_file(self):
        with salt.utils.files.fopen(self.module_file, "w") as fh:
            fh.write(self.module_content)
            fh.flush()
            os.fsync(fh.fileno())

    def _byte_compile(self):
        compileall.compile_file(self.module_file, quiet=1, optimize=0)
        compileall.compile_file(self.module_file, quiet=1, optimize=1)
        compileall.compile_file(self.module_file, quiet=1, optimize=2)

    def _test_optimization_order(self, order):
        self._write_module_file()
        self._byte_compile()

        # Clean up the original file so that we can be assured we're only
        # loading the byte-compiled files(s).
        os.remove(self.module_file)

        self.loader = self._get_loader(order)
        filename = self._get_module_filename()
        basename = os.path.basename(filename)
        assert basename == self._expected(order[0]), basename

        # Remove the file and make a new loader. We should now load the
        # byte-compiled file with an optimization level matching the 2nd
        # element of the order list.
        os.remove(filename)
        self.loader = self._get_loader(order)
        filename = self._get_module_filename()
        basename = os.path.basename(filename)
        assert basename == self._expected(order[1]), basename

        # Remove the file and make a new loader. We should now load the
        # byte-compiled file with an optimization level matching the 3rd
        # element of the order list.
        os.remove(filename)
        self.loader = self._get_loader(order)
        filename = self._get_module_filename()
        basename = os.path.basename(filename)
        assert basename == self._expected(order[2]), basename

    def test_optimization_order(self):
        """
        Test the optimization_order config param
        """
        self._test_optimization_order([0, 1, 2])
        self._test_optimization_order([0, 2, 1])
        self._test_optimization_order([1, 2, 0])
        self._test_optimization_order([1, 0, 2])
        self._test_optimization_order([2, 0, 1])
        self._test_optimization_order([2, 1, 0])

    def test_load_source_file(self):
        """
        Make sure that .py files are preferred over .pyc files
        """
        self._write_module_file()
        self._byte_compile()
        self.loader = self._get_loader()
        filename = self._get_module_filename()
        basename = os.path.basename(filename)
        expected = "lazyloadertest.py"
        assert basename == expected, basename


class LoaderLoadCachedGrainsTest(TestCase):
    """
    Test how the loader works with cached grains
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

    def setUp(self):
        self.cache_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.cache_dir, ignore_errors=True)

        self.opts["cachedir"] = self.cache_dir
        self.opts["grains_cache"] = True
        self.opts["grains"] = salt.loader.grains(self.opts)

    @pytest.mark.slow_test
    def test_osrelease_info_has_correct_type(self):
        """
        Make sure osrelease_info is tuple after caching
        """
        grains = salt.loader.grains(self.opts)
        osrelease_info = grains["osrelease_info"]
        assert isinstance(osrelease_info, tuple), osrelease_info

    def test_deserialization_error_caught(self):
        """
        Test that _load_cached_grains catches SaltDeserializationError.

        When msgpack encounters corrupted cache data, it raises
        SaltDeserializationError. This should be caught and return None
        so grains can be regenerated.
        """
        import time

        import salt.exceptions

        with patch("os.path.isfile", return_value=True), patch(
            "os.path.getmtime", return_value=time.time()
        ), patch("salt.utils.files.fopen"), patch(
            "salt.payload.load",
            side_effect=salt.exceptions.SaltDeserializationError(
                "Could not deserialize msgpack message"
            ),
        ):
            result = salt.loader._load_cached_grains(self.opts, "/fake/path")
            # Should return None, not raise exception
            self.assertIsNone(result)


class LazyLoaderRefreshFileMappingTest(TestCase):
    """
    Test that _refresh_file_mapping is called using acquiring LazyLoader._lock
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.minion_config(None)
        cls.opts["grains"] = salt.loader.grains(cls.opts)
        cls.utils = salt.loader.utils(copy.deepcopy(cls.opts))
        cls.proxy = salt.loader.proxy(cls.opts)
        cls.funcs = salt.loader.minion_mods(cls.opts, utils=cls.utils, proxy=cls.proxy)

    def setUp(self):
        class LazyLoaderMock(salt.loader.LazyLoader):
            pass

        self.LOADER_CLASS = LazyLoaderMock

    def __init_loader(self):
        return self.LOADER_CLASS(
            salt.loader._module_dirs(copy.deepcopy(self.opts), "modules", "module"),
            copy.deepcopy(self.opts),
            tag="module",
            pack={
                "__utils__": self.utils,
                "__salt__": self.funcs,
                "__proxy__": self.proxy,
            },
        )

    @classmethod
    def tearDownClass(cls):
        del cls.opts
        del cls.utils
        del cls.funcs
        del cls.proxy

    def test_lazyloader_refresh_file_mapping_called_with_lock_at___init__(self):
        func_mock = MagicMock()
        lock_mock = MagicMock()
        lock_mock.__enter__ = MagicMock()
        self.LOADER_CLASS._refresh_file_mapping = func_mock
        with patch.object(
            self.LOADER_CLASS, "_get_lock", MagicMock(return_value=lock_mock)
        ):
            loader = self.__init_loader()
        lock_mock.__enter__.assert_called()
        func_mock.assert_called()
        assert len(func_mock.call_args_list) == len(lock_mock.__enter__.call_args_list)
        del loader

    def test_lazyloader_zip_modules(self):
        self.opts["enable_zip_modules"] = True
        try:
            loader = self.__init_loader()
            assert ".zip" in loader.suffix_map
            assert ".zip" in loader.suffix_order
        finally:
            self.opts["enable_zip_modules"] = False
        loader = self.__init_loader()
        assert ".zip" not in loader.suffix_map
        assert ".zip" not in loader.suffix_order

    def test_lazyloader_pyx_modules(self):
        self.opts["cython_enable"] = True
        try:
            loader = self.__init_loader()
            # Don't assert if the current environment has no pyximport
            if salt.loader.lazy.pyximport is not None:
                assert ".pyx" in loader.suffix_map
                assert ".pyx" in loader.suffix_order
        finally:
            self.opts["cython_enable"] = False
        loader = self.__init_loader()
        assert ".pyx" not in loader.suffix_map
        assert ".pyx" not in loader.suffix_order



import logging
import os
import pathlib

import attr
import pytest
import requests

import salt.modules.aptpkg
import salt.utils.files
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.timeout_unless_on_windows(120),
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.slow_test,
]


@pytest.fixture
def pkgrepo(states, grains):
    if grains["os_family"] != "Debian":
        raise pytest.skip.Exception(
            "Test only for debian based platforms", _use_item_location=True
        )
    sources = pathlib.Path("/etc/apt/sources.list")
    if not sources.exists():
        sources.touch()
    return states.pkgrepo


@pytest.mark.requires_salt_states("pkgrepo.managed")
def test_adding_repo_file(pkgrepo, repo_uri, tmp_path):
    """
    test adding a repo file using pkgrepo.managed
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
    repo_file = str(tmp_path / "stable-binary.list")
    repo_content = f"deb{signedby} {repo_uri} stable main"
    ret = pkgrepo.managed(
        name=repo_content, file=repo_file, clean_file=True, aptkey=aptkey
    )
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read().strip()
    assert file_content == repo_content


@pytest.mark.requires_salt_states("pkgrepo.managed")
def test_adding_repo_file_arch(pkgrepo, repo_uri, tmp_path, subtests):
    """
    test adding a repo file using pkgrepo.managed
    and setting architecture
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg"
    repo_file = str(tmp_path / "stable-binary.list")
    repo_content = f"deb [arch=amd64{signedby} ] {repo_uri} stable main"
    ret = pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read().strip()
        assert file_content == f"deb [arch=amd64{signedby}] {repo_uri} stable main"
    with subtests.test("With multiple archs"):
        repo_content = f"deb [arch=amd64,i386{signedby}  ] {repo_uri} stable main"
        pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
        with salt.utils.files.fopen(repo_file, "r") as fp:
            file_content = fp.read().strip()
            assert (
                file_content
                == f"deb [arch=amd64,i386{signedby}] {repo_uri} stable main"
            )


@pytest.mark.requires_salt_states("pkgrepo.managed")
def test_adding_repo_file_cdrom(pkgrepo, tmp_path):
    """
    test adding a repo file using pkgrepo.managed
    The issue is that CDROM installs often have [] in the line, and we
    should still add the repo even though it's not setting arch(for example)
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
    repo_file = str(tmp_path / "cdrom.list")
    repo_content = f"deb{signedby} cdrom:[Debian GNU/Linux 11.4.0 _Bullseye_ - Official amd64 NETINST 20220709-10:31]/ stable main"
    ret = pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read().strip()
        assert (
            file_content
            == f"deb{signedby} cdrom:[Debian GNU/Linux 11.4.0 _Bullseye_ - Official amd64 NETINST 20220709-10:31]/ stable main"
        )


@pytest.fixture
def ubuntu_state_tree(state_tree, grains):
    if grains["os"] != "Ubuntu":
        pytest.skip(
            "Test only applicable to Ubuntu, not '{}'".format(grains["osfinger"])
        )
    managed_sls_contents = """
    {% set codename = grains['oscodename'] %}
    {% set ubuntu_repos = [] %}
    {% set beta = grains['oscodename'] in ['xenial', 'bionic', 'eoan', 'focal', 'groovy'] %}
    {% set backports = grains['oscodename'] in ['xenial', 'bionic', 'eoan', 'focal'] %}

    {%- if beta %}{%- do ubuntu_repos.append('firefox-beta') %}
    firefox-beta:
      pkgrepo.managed:
        - name: deb http://ppa.launchpad.net/mozillateam/firefox-next/ubuntu {{ codename }} main
        - dist: {{ codename }}
        - file: /etc/apt/sources.list.d/firefox-beta.list
        - keyid: CE49EC21
        - keyserver: keyserver.ubuntu.com
    {%- endif %}

    {%- if backports %}{%- do ubuntu_repos.append('kubuntu-ppa') %}
    kubuntu-ppa:
      pkgrepo.managed:
        - ppa: kubuntu-ppa/backports
    {%- endif %}

    pkgrepo-deps:
      pkg.installed:
        - pkgs:
          - python3-apt
          - software-properties-common
    {%- for repo in ubuntu_repos -%}
    {% if loop.first %}
        - require_in:{%- endif %}
          - pkgrepo: {{ repo }}
    {%- endfor %}
    """
    absent_sls_contents = """
    firefox-beta:
      pkgrepo.absent:
        - name: deb http://ppa.launchpad.net/mozillateam/firefox-next/ubuntu {{ grains['oscodename'] }} main

    kubuntu-ppa:
      pkgrepo.absent:
        - ppa: kubuntu-ppa/backports
    """
    managed_state_file = pytest.helpers.temp_file(
        "pkgrepo/managed.sls", managed_sls_contents, state_tree
    )
    absent_state_file = pytest.helpers.temp_file(
        "pkgrepo/absent.sls", absent_sls_contents, state_tree
    )
    try:
        with managed_state_file, absent_state_file:
            yield
    finally:
        for pathstr in ("/etc/apt/sources.list.d/firefox-beta.list",):
            path = pathlib.Path(pathstr)
            if path.exists():
                path.unlink()


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_pkgrepo_managed_absent(modules, ubuntu_state_tree, subtests):
    """
    Test adding a repo with the system aptsources package
    """

    add_repo_test_passed = False

    with subtests.test("Add Repo"):
        ret = modules.state.sls("pkgrepo.managed")
        assert ret.failed is False
        for state in ret:
            assert state.result is True
        add_repo_test_passed = True

    with subtests.test("Remove Repo"):
        if add_repo_test_passed is False:
            pytest.skip("Adding the repo failed. Skipping.")
        ret = modules.state.sls("pkgrepo.absent")
        assert ret.failed is False
        for state in ret:
            assert state.result is True


@pytest.fixture
def multiple_comps_repo_file_caconical(grains):
    if grains["os"] != "Ubuntu":
        pytest.skip(
            "Test only applicable to Ubuntu, not '{}'".format(grains["osfinger"])
        )
    repo_file_path = "/etc/apt/sources.list.d/99-salt-canonical-ubuntu.list"
    try:
        yield repo_file_path
    finally:
        try:
            os.unlink(repo_file_path)
        except OSError:
            pass


@pytest.fixture
def multiple_comps_repo_file_backports(grains):
    if grains["os"] != "Ubuntu":
        pytest.skip(
            "Test only applicable to Ubuntu, not '{}'".format(grains["osfinger"])
        )
    repo_file_path = (
        "/etc/apt/sources.list.d/99-salt-archive-ubuntu-focal-backports.list"
    )
    try:
        yield repo_file_path
    finally:
        try:
            os.unlink(repo_file_path)
        except OSError:
            pass


@pytest.fixture
def multiple_comps_state_tree(
    multiple_comps_repo_file_caconical, multiple_comps_repo_file_backports, state_tree
):
    sls_contents = """
    ubuntu-backports:
      pkgrepo.managed:
        - name: 'deb http://fi.archive.ubuntu.com/ubuntu focal-backports'
        - comps: main, restricted, universe, multiverse
        - refresh: false
        - disabled: false
        - clean_file: true
        - file: {}
        - require_in:
          - pkgrepo: canonical-ubuntu

    canonical-ubuntu:
      pkgrepo.managed:
        - name: 'deb http://archive.canonical.com/ubuntu {{{{ salt['grains.get']('oscodename') }}}}'
        - comps: partner
        - refresh: false
        - disabled: false
        - clean_file: true
        - file: {}
    """.format(
        multiple_comps_repo_file_backports,
        multiple_comps_repo_file_caconical,
    )
    with pytest.helpers.temp_file("multiple-comps-repos.sls", sls_contents, state_tree):
        yield


def test_managed_multiple_comps(modules, multiple_comps_state_tree):
    # On the first run, we must have changes
    ret = modules.state.sls("multiple-comps-repos")
    assert ret.failed is False
    for state in ret:
        assert state.result is True
        assert state.changes

    # On the second run though, we shouldn't have changes made
    ret = modules.state.sls("multiple-comps-repos")
    assert ret.failed is False
    for state in ret:
        assert state.result is True
        assert not state.changes


@pytest.fixture
def sources_list_file():
    fn_ = salt.utils.files.mkstemp(dir="/etc/apt/sources.list.d", suffix=".list")
    try:
        yield fn_
    finally:
        try:
            os.remove(fn_)
        except OSError:
            pass


def test_pkgrepo_with_architectures(pkgrepo, grains, sources_list_file, subtests):
    """
    Test managing a repo with architectures specified
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = "signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg"
    name = "deb {{arch}}http://foo.com/bar/latest {oscodename} main".format(
        oscodename=grains["oscodename"]
    )

    def _get_arch(arch):
        if signedby:
            return f"[arch={arch} {signedby}] " if arch else f"[{signedby}] "
        else:
            return f"[arch={arch}] " if arch else ""

    def _run(arch=None, test=False):
        return pkgrepo.managed(
            name=name.format(arch=_get_arch(arch)),
            file=sources_list_file,
            refresh=False,
            test=test,
        )

    with subtests.test("test=True"):
        # Run with test=True
        ret = _run(test=True)
        assert ret.changes == {"repo": name.format(arch=_get_arch(""))}
        assert "would be" in ret.comment
        assert ret.result is None

    with subtests.test("test=False"):
        # Run for real
        ret = _run()
        assert ret.changes == {"repo": name.format(arch=_get_arch(""))}
        assert ret.comment.startswith("Configured")
        assert ret.result is True

    with subtests.test("test=True repeat"):
        # Run again with test=True, should exit with no changes and a True
        # result.
        ret = _run(test=True)
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True

    with subtests.test("test=False repeat"):
        # Run for real again, results should be the same as above (i.e. we
        # should never get to the point where we exit with a None result).
        ret = _run()
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True

    expected_changes = {
        "line": {
            "new": name.format(arch=_get_arch("amd64")),
            "old": name.format(arch=_get_arch("")),
        },
        "architectures": {"new": ["amd64"], "old": []},
    }
    with subtests.test("test=True arch=amd64"):
        # Run with test=True and the architecture set. We should get a None
        # result with some expected changes.
        ret = _run(arch="amd64", test=True)
        assert ret.changes == expected_changes
        assert "would be" in ret.comment
        assert ret.result is None

    with subtests.test("test=False arch=amd64"):
        # Run for real, with the architecture set. We should get a True
        # result with the same changes.
        ret = _run(arch="amd64")
        assert ret.changes == expected_changes
        assert ret.comment.startswith("Configured")
        assert ret.result is True

    with subtests.test("test=True arch=amd64 repeat"):
        # Run again with test=True, should exit with no changes and a True
        # result.
        ret = _run(arch="amd64", test=True)
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True

    with subtests.test("test=False arch=amd64 repeat"):
        # Run for real again, results should be the same as above (i.e. we
        # should never get to the point where we exit with a None result).
        ret = _run(arch="amd64")
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True

    expected_changes = {
        "line": {
            "new": name.format(arch=_get_arch("")),
            "old": name.format(arch=_get_arch("amd64")),
        },
        "architectures": {"new": [], "old": ["amd64"]},
    }

    with subtests.test("test=True arch=None"):
        # Run with test=True and the architecture set back to the original
        # value. We should get a None result with some expected changes.
        ret = _run(test=True)
        assert ret.changes == expected_changes
        assert "would be" in ret.comment
        assert ret.result is None

    with subtests.test("test=False arch=None"):
        # Run for real, with the architecture set. We should get a True
        # result with the same changes.
        ret = _run()
        assert ret.changes == expected_changes
        assert ret.comment.startswith("Configured")
        assert ret.result is True

    with subtests.test("test=True arch=None repeat"):
        # Run again with test=True, should exit with no changes and a True
        # result.
        ret = _run(test=True)
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True

    with subtests.test("test=False arch=None repeat"):
        # Run for real again, results should be the same as above (i.e. we
        # should never get to the point where we exit with a None result).
        ret = _run()
        assert not ret.changes
        assert "already" in ret.comment
        assert ret.result is True


@pytest.fixture(scope="module")
def repo_uri():
    yield "http://www.deb-multimedia.org"


@pytest.fixture
def trailing_slash_repo_file(grains):
    if grains["os_family"] != "Debian":
        pytest.skip(
            "Test only applicable to Debian flavors, not '{}'".format(
                grains["osfinger"]
            )
        )
    repo_file_path = "/etc/apt/sources.list.d/trailing-slash.list"
    try:
        yield repo_file_path
    finally:
        try:
            os.unlink(repo_file_path)
        except OSError:
            pass


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_repo_present_absent_trailing_slash_uri(
    pkgrepo, repo_uri, trailing_slash_repo_file
):
    """
    test adding and then removing a repo with a trailing slash in the uri
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
    # with the trailing slash
    repo_content = f"deb{signedby} {repo_uri}/ stable main"
    # initial creation
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False, clean_file=True
    )
    with salt.utils.files.fopen(trailing_slash_repo_file, "r") as fp:
        file_content = fp.read()
    assert file_content.strip() == f"deb{signedby} {repo_uri}/ stable main"
    assert ret.changes
    # no changes
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False
    )
    assert not ret.changes
    # absent
    ret = pkgrepo.absent(name=repo_content)
    assert ret.result


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_repo_present_absent_no_trailing_slash_uri(
    pkgrepo, repo_uri, trailing_slash_repo_file
):
    """
    test adding a repo with a trailing slash in the uri
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
    # without the trailing slash
    repo_content = f"deb{signedby} {repo_uri} stable main"
    # initial creation
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False, clean_file=True
    )
    with salt.utils.files.fopen(trailing_slash_repo_file, "r") as fp:
        file_content = fp.read()
    assert file_content.strip() == repo_content
    assert ret.changes
    # no changes
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False
    )
    assert not ret.changes
    # absent
    ret = pkgrepo.absent(name=repo_content)
    assert ret.result


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_repo_present_absent_no_trailing_slash_uri_add_slash(
    pkgrepo, repo_uri, trailing_slash_repo_file
):
    """
    test adding a repo without a trailing slash, and then running it
    again with a trailing slash.
    """
    aptkey = True if salt.utils.path.which("apt-key") else False
    signedby = ""
    if not aptkey:
        signedby = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
    # without the trailing slash
    repo_content = f"deb{signedby} {repo_uri} stable main"
    # initial creation
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False, clean_file=True
    )
    with salt.utils.files.fopen(trailing_slash_repo_file, "r") as fp:
        file_content = fp.read()
    assert file_content.strip() == repo_content
    assert ret.changes
    # now add a trailing slash in the name
    repo_content = f"deb{signedby} {repo_uri}/ stable main"
    ret = pkgrepo.managed(
        name=repo_content, file=trailing_slash_repo_file, refresh=False
    )
    with salt.utils.files.fopen(trailing_slash_repo_file, "r") as fp:
        file_content = fp.read()
    assert file_content.strip() == repo_content
    # absent
    ret = pkgrepo.absent(name=repo_content)
    assert ret.result


@pytest.mark.requires_salt_states("pkgrepo.absent")
def test_repo_absent_existing_repo_trailing_slash_uri(
    pkgrepo, repo_uri, subtests, trailing_slash_repo_file
):
    """
    Test pkgrepo.absent with a URI containing a trailing slash

    This test is different from test_repo_present_absent_trailing_slash_uri.
    That test first does a pkgrepo.managed with a URI containing a trailing
    slash. Since pkgrepo.managed normalizes the URI by removing the trailing
    slash, the resulting repo file created by Salt does not contain one. This
    tests the case where Salt is asked to remove an existing repo with a
    trailing slash in the repo URI.

    See https://github.com/saltstack/salt/issues/64286
    """
    repo_file = pathlib.Path(trailing_slash_repo_file)
    repo_content = f"deb [arch=amd64] {repo_uri}/ stable main"

    with subtests.test("Remove repo with trailing slash in URI"):
        # Write contents to file with trailing slash in URI
        repo_file.write_text(f"{repo_content}\n", encoding="utf-8")
        # Perform and validate removal
        ret = pkgrepo.absent(name=repo_content)
        assert ret.result
        assert ret.changes
        assert not repo_file.exists()
        # A second run of the pkgrepo.absent state should be a no-op (i.e. no changes)
        ret = pkgrepo.absent(name=repo_content)
        assert ret.result
        assert not ret.changes
        assert not repo_file.exists()

    with subtests.test("URI match with mismatched arch"):
        # Create a repo file that matches the URI but contains no architecture.
        # This should not be identified as a match for repo_content, and thus
        # the result of a state should be a no-op.
        repo_file.write_text(f"deb {repo_uri} stable main\n", encoding="utf-8")
        # Since this was a no-op, the state should have succeeded, made no
        # changes, and left the repo file in place.
        ret = pkgrepo.absent(name=repo_content)
        assert ret.result
        assert not ret.changes
        assert repo_file.exists()


@attr.s(kw_only=True)
class Repo:
    key_root = attr.ib(default=pathlib.Path("/usr", "share", "keyrings"))
    signedby = attr.ib(default=False)
    grains = attr.ib()
    fullname = attr.ib()
    alt_repo = attr.ib(init=False)
    key_file = attr.ib()
    sources_list_file = attr.ib()
    repo_file = attr.ib()
    repo_url = attr.ib()
    repo_content = attr.ib()
    key_url = attr.ib()

    @fullname.default
    def _default_fullname(self):
        return self.grains["osfullname"].lower().split()[0]

    @alt_repo.default
    def _default_alt_repo(self):
        """
        Use an alternative repo, packages do not exist for the OS on
        packages.broadcom.com
        """
        if (
            self.grains["osfullname"] == "Ubuntu"
            and self.grains["osrelease"] == "22.04"
            or "Debian" in self.grains["osfullname"]
            and self.grains["osrelease"] == "12"
            # only need to use alt repo until
            # we release Debian 12 salt packages
        ):
            return True
        return False

    @key_file.default
    def _default_key_file(self):
        key_file = self.key_root / "salt-archive-keyring.gpg"
        if self.alt_repo:
            key_file = self.key_root / "elasticsearch-keyring.gpg"
        key_file.parent.mkdir(exist_ok=True)
        assert not key_file.is_file()
        return key_file

    @repo_file.default
    def _default_repo_file(self):
        return self.sources_list_file

    @repo_url.default
    def _default_repo_url(self):
        return "https://packages.broadcom.com/artifactory/saltproject-deb/"

    @repo_content.default
    def _default_repo_content(self):
        if self.alt_repo:
            opts = " "
            if self.signedby:
                opts = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg] "
            repo_content = (
                "deb{}https://artifacts.elastic.co/packages/8.x/apt stable main".format(
                    opts
                )
            )
        else:
            opts = "[arch={arch}]".format(arch=self.grains["osarch"])
            if self.signedby:
                opts = "[arch={arch} signed-by=/usr/share/keyrings/salt-archive-keyring.gpg]".format(
                    arch=self.grains["osarch"]
                )
            repo_content = (
                f"deb {opts} {self.repo_url} {self.grains['oscodename']} main"
            )
        return repo_content

    @key_url.default
    def _default_key_url(self):
        key_url = f"{self.repo_url}/salt-archive-keyring.gpg"
        if self.alt_repo:
            key_url = "https://artifacts.elastic.co/GPG-KEY-elasticsearch"
        return key_url

    @property
    def exists(self):
        """
        Return True if the repository path exists.
        """
        response = requests.head(self.key_url, timeout=30)
        return response.status_code == 200


@pytest.fixture
def repo(request, grains, sources_list_file):
    signedby = False
    if "signedby" in request.node.name:
        signedby = True
    repo = Repo(grains=grains, sources_list_file=sources_list_file, signedby=signedby)
    if not repo.exists:
        pytest.skip(f"The repo url '{repo.repo_url}' does not exist")
    try:
        yield repo
    finally:
        for key in [repo.key_file, repo.key_file.parent / "salt-alt-key.gpg"]:
            if key.is_file():
                key.unlink()


def test_adding_repo_file_signedby(pkgrepo, states, repo, subtests):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby
    """

    def _run(test=False):
        return states.pkgrepo.managed(
            name=repo.repo_content,
            file=str(repo.repo_file),
            clean_file=True,
            signedby=str(repo.key_file),
            key_url=repo.key_url,
            aptkey=False,
            test=test,
        )

    ret = _run()
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert file_content.strip() == repo.repo_content
    assert repo.key_file.is_file()
    assert repo.repo_content in ret.comment
    with subtests.test("test=True"):
        ret = _run(test=True)
        assert ret.changes == {}


def test_adding_repo_file_signedby_invalid_name(pkgrepo, states, repo):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby and the name is invalid.
    Ensure we raise an error.
    """

    default_sources = pathlib.Path("/etc", "apt", "sources.list")
    with salt.utils.files.fopen(default_sources, "r") as fp:
        pre_file_content = fp.read()

    ret = states.pkgrepo.managed(
        name=repo.repo_content.strip("deb"),
        file=str(repo.repo_file),
        clean_file=True,
        signedby=str(repo.key_file),
        key_url=repo.key_url,
        aptkey=False,
        test=False,
    )

    assert "Failed to configure repo" in ret.comment
    assert "This must be the complete repo entry" in ret.comment
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert not file_content

    with salt.utils.files.fopen(default_sources, "r") as fp:
        post_file_content = fp.read()
    assert pre_file_content == post_file_content


def test_adding_repo_file_signedby_keyserver(pkgrepo, states, repo):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby with a keyserver
    """
    ret = states.pkgrepo.managed(
        name=repo.repo_content,
        file=str(repo.repo_file),
        clean_file=True,
        signedby=str(repo.key_file),
        keyserver="keyserver.ubuntu.com",
        keyid="0E08A149DE57BFBE",
        aptkey=False,
    )
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert file_content.strip() == repo.repo_content
    assert repo.key_file.is_file()


def test_adding_repo_file_keyserver_key_url(pkgrepo, states, repo):
    """
    Test adding a repo file using pkgrepo.managed
    and a key_url.
    """
    ret = states.pkgrepo.managed(
        name=repo.repo_content,
        file=str(repo.repo_file),
        clean_file=True,
        keyserver="keyserver.ubuntu.com",
        key_url=repo.key_url,
    )
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert file_content.strip() == repo.repo_content


def test_adding_repo_file_signedby_alt_file(pkgrepo, states, repo):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby and then running again with
    different key path.
    """
    ret = states.pkgrepo.managed(
        name=repo.repo_content,
        file=str(repo.repo_file),
        clean_file=True,
        key_url=repo.key_url,
        aptkey=False,
    )
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert file_content.strip() == repo.repo_content
    assert repo.key_file.is_file()
    assert repo.repo_content in ret.comment

    key_file = repo.key_file.parent / "salt-alt-key.gpg"
    repo_content = "deb [arch=amd64 signed-by={}] https://packages.broadcom.com/artifactory/saltproject-deb/ buster main".format(
        str(key_file)
    )
    ret = states.pkgrepo.managed(
        name=repo_content,
        file=str(repo.repo_file),
        clean_file=True,
        key_url=repo.key_url,
        aptkey=False,
    )
    with salt.utils.files.fopen(str(repo.repo_file), "r") as fp:
        file_content = fp.read()
        assert file_content.strip() == repo_content
        assert file_content.endswith("\n")
    assert key_file.is_file()
    assert repo_content in ret.comment


def test_adding_repo_file_signedby_fail_key_keyid(
    pkgrepo, states, repo, subtests, modules
):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby and keyid when adding the key fails
    an error is returned
    """

    def _run(test=False):
        return states.pkgrepo.managed(
            name=repo.repo_content,
            file=str(repo.repo_file),
            clean_file=True,
            signedby=str(repo.key_file),
            keyid="10857FFDD3F91EAE577A21D664CBBC8173D76B3F1",
            keyserver="keyserver.ubuntu.com",
            aptkey=False,
            test=test,
            keydir="/tmp/test",
        )

    ret = _run()
    assert "Failed to configure repo" in ret.comment
    assert "Could not add key" in ret.comment


def test_adding_repo_file_signedby_fail_key_keyurl(
    pkgrepo, states, repo, subtests, modules
):
    """
    Test adding a repo file using pkgrepo.managed
    and setting signedby and keyurl when adding the key fails
    an error is returned
    """

    def _run(test=False):
        with patch(
            "salt.utils.path.which", MagicMock(side_effect=[True, True, False, False])
        ):
            return states.pkgrepo.managed(
                name=repo.repo_content,
                file=str(repo.repo_file),
                clean_file=True,
                key_url="https://packages.broadcom.com/artifactory/api/security/keypair/SaltProjectKey/public",
                aptkey=False,
            )

    ret = _run()
    assert "Failed to configure repo" in ret.comment
    assert "Could not add key" in ret.comment


@pytest.fixture
def deb822_repo_file(grains):
    if grains["os_family"] != "Debian":
        pytest.skip(
            "Test only applicable to Debian flavors, not '{}'".format(
                grains["osfinger"]
            )
        )
    repo_file_path = "/etc/apt/sources.list.d/deb822-test.sources"
    try:
        yield repo_file_path
    finally:
        try:
            os.unlink(repo_file_path)
        except OSError:
            pass


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_repo_present_absent_deb822_repo_file(
    pkgrepo, grains, deb822_repo_file, subtests
):
    """
    test adding and managing a deb822 repo.
    """
    codename = grains["oscodename"]
    repo_uri_main = "http://ftp.es.debian.org/debian"
    repo_uri_alt = "http://ftp.cz.debian.org/debian"
    aptkey = True if salt.utils.path.which("apt-key") else False
    ext_attrs = ""
    expected_ext_attrs = ""
    if not aptkey:
        ext_attrs = " [signed-by=/usr/share/keyrings/elasticsearch-keyring.gpg]"
        expected_ext_attrs = (
            "\nSigned-By: /usr/share/keyrings/elasticsearch-keyring.gpg"
        )

    # without the trailing slash
    repo_content = f"deb{ext_attrs} {repo_uri_main} {codename} main"
    expected_content = f"""Types: deb
URIs: {repo_uri_main}
Suites: {codename}
Components: main{expected_ext_attrs}"""
    with subtests.test("Create new deb822 repo source"):
        # initial creation
        ret = pkgrepo.managed(
            name=repo_content, file=deb822_repo_file, refresh=False, clean_file=True
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # use trailing slash in the URI and add extra suites
    repo_content = f"deb{ext_attrs} {repo_uri_main}/ {codename} main"
    expected_content = f"""Types: deb
URIs: {repo_uri_main}/
Suites: {codename} {codename}-updates {codename}-backports
Components: main{expected_ext_attrs}"""
    with subtests.test("Add suites to deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content,
            file=deb822_repo_file,
            refresh=False,
            suites=[codename, f"{codename}-updates", f"{codename}-backports"],
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # use trailing slash in the URI and add extra components
    repo_content = f"deb{ext_attrs} {repo_uri_main}/ {codename} main"
    expected_content = f"""Types: deb
URIs: {repo_uri_main}/
Suites: {codename} {codename}-updates {codename}-backports
Components: main contrib{expected_ext_attrs}"""
    with subtests.test("Add comps to deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content,
            file=deb822_repo_file,
            refresh=False,
            comps="main,contrib",
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # use trailing slash in the URI and add extra type
    repo_content = f"deb{ext_attrs} {repo_uri_main}/ {codename} main contrib"
    expected_content = f"""Types: deb deb-src
URIs: {repo_uri_main}/
Suites: {codename} {codename}-updates {codename}-backports
Components: main contrib{expected_ext_attrs}"""
    with subtests.test("Add type to deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content,
            file=deb822_repo_file,
            refresh=False,
            types=["deb", "deb-src"],
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # do not use trailing slash in the URI and add extra URI
    repo_content = f"deb{ext_attrs} {repo_uri_main} {codename} main contrib"
    expected_content = f"""Types: deb deb-src
URIs: {repo_uri_main} {repo_uri_alt}
Suites: {codename} {codename}-updates {codename}-backports
Components: main contrib{expected_ext_attrs}"""
    with subtests.test("Add extra URI to deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content,
            file=deb822_repo_file,
            refresh=False,
            uris=[repo_uri_main, repo_uri_alt],
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # do not use trailing slash in the URI and remove suites
    repo_content = f"deb{ext_attrs} {repo_uri_main} {codename} main contrib"
    expected_content = f"""Types: deb deb-src
URIs: {repo_uri_main} {repo_uri_alt}
Suites: {codename}
Components: main contrib{expected_ext_attrs}"""
    with subtests.test("Remove suites from deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content,
            file=deb822_repo_file,
            refresh=False,
            suites=[codename],
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # Add one more repo source to the existing source file
    repo_uri_main_sec = "http://ftp.cz.debian.org/debian-security"
    repo_uri_alt_sec = "http://ftp.es.debian.org/debian-security"
    repo_content_sec = (
        f"deb{ext_attrs} {repo_uri_main_sec} {codename}-security main updates"
    )
    expected_content = f"""Types: deb deb-src
URIs: {repo_uri_main} {repo_uri_alt}
Suites: {codename}
Components: main contrib{expected_ext_attrs}

Types: deb
URIs: {repo_uri_main_sec} {repo_uri_alt_sec}
Suites: {codename}-security
Components: main updates{expected_ext_attrs}"""
    with subtests.test("Add extra deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content_sec,
            file=deb822_repo_file,
            refresh=False,
            uris=[repo_uri_main_sec, repo_uri_alt_sec],
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # Disable repo source and leave just alternative URI
    repo_content_sec = (
        f"deb{ext_attrs} {repo_uri_main_sec} {codename}-security main updates"
    )
    expected_content = f"""Types: deb deb-src
URIs: {repo_uri_main} {repo_uri_alt}
Suites: {codename}
Components: main contrib{expected_ext_attrs}

Types: deb
URIs: {repo_uri_alt_sec}
Suites: {codename}-security
Components: main updates{expected_ext_attrs}
Enabled: no"""
    with subtests.test("Disable extra deb822 repo source"):
        ret = pkgrepo.managed(
            name=repo_content_sec,
            file=deb822_repo_file,
            refresh=False,
            uris=[repo_uri_alt_sec],
            disabled=True,
        )
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert file_content.strip() == expected_content
        assert ret.changes

    # Remove repo source
    expected_content = f"""Types: deb
URIs: {repo_uri_alt_sec}
Suites: {codename}-security
Components: main updates{expected_ext_attrs}
Enabled: no"""
    with subtests.test("Remove deb822 repo source"):
        ret = pkgrepo.absent(name=repo_content)
        with salt.utils.files.fopen(deb822_repo_file, "r") as fp:
            file_content = fp.read()
        assert ret.result
        assert file_content.strip() == expected_content



"""
    tests.integration.shell.master_tops
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

from tests.support.case import ShellCase


@pytest.mark.windows_whitelisted
class MasterTopsTest(ShellCase):

    _call_binary_ = "salt"

    @pytest.mark.slow_test
    def test_custom_tops_gets_utilized(self):
        resp = self.run_call("state.show_top")
        self.assertTrue(any("master_tops_test" in _x for _x in resp))



"""
Manage and query udev info

.. versionadded:: 2015.8.0

"""

import logging

import salt.modules.cmdmod
import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

__salt__ = {
    "cmd.run_all": salt.modules.cmdmod.run_all,
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work when udevadm is installed.
    """
    return salt.utils.path.which_bin(["udevadm"]) is not None


def _parse_udevadm_info(udev_info):
    """
    Parse the info returned by udevadm command.
    """
    devices = []
    dev = {}

    for line in (line.strip() for line in udev_info.splitlines()):
        if line:
            line = line.split(":", 1)
            if len(line) != 2:
                continue
            query, data = line
            if query == "E":
                if query not in dev:
                    dev[query] = {}
                key, val = data.strip().split("=", 1)

                try:
                    val = int(val)
                except ValueError:
                    try:
                        val = float(val)
                    except ValueError:
                        pass  # Quiet, this is not a number.

                dev[query][key] = val
            else:
                if query not in dev:
                    dev[query] = []
                dev[query].append(data.strip())
        else:
            if dev:
                devices.append(_normalize_info(dev))
                dev = {}
    if dev:
        _normalize_info(dev)
        devices.append(_normalize_info(dev))

    return devices


def _normalize_info(dev):
    """
    Replace list with only one element to the value of the element.

    :param dev:
    :return:
    """
    for sect, val in dev.items():
        if len(val) == 1:
            dev[sect] = val[0]

    return dev


def info(dev):
    """
    Extract all info delivered by udevadm

    CLI Example:

    .. code-block:: bash

        salt '*' udev.info /dev/sda
        salt '*' udev.info /sys/class/net/eth0
    """
    if "sys" in dev:
        qtype = "path"
    else:
        qtype = "name"

    cmd = f"udevadm info --export --query=all --{qtype}={dev}"
    udev_result = __salt__["cmd.run_all"](cmd, output_loglevel="quiet")

    if udev_result["retcode"] != 0:
        raise CommandExecutionError(udev_result["stderr"])

    return _parse_udevadm_info(udev_result["stdout"])[0]


def env(dev):
    """
    Return all environment variables udev has for dev

    CLI Example:

    .. code-block:: bash

        salt '*' udev.env /dev/sda
        salt '*' udev.env /sys/class/net/eth0
    """
    return info(dev).get("E", None)


def name(dev):
    """
    Return the actual dev name(s?) according to udev for dev

    CLI Example:

    .. code-block:: bash

        salt '*' udev.dev /dev/sda
        salt '*' udev.dev /sys/class/net/eth0
    """
    return info(dev).get("N", None)


def path(dev):
    """
    Return the physical device path(s?) according to udev for dev

    CLI Example:

    .. code-block:: bash

        salt '*' udev.path /dev/sda
        salt '*' udev.path /sys/class/net/eth0
    """
    return info(dev).get("P", None)


def links(dev):
    """
    Return all udev-created device symlinks

    CLI Example:

    .. code-block:: bash

        salt '*' udev.links /dev/sda
        salt '*' udev.links /sys/class/net/eth0
    """
    return info(dev).get("S", None)


def _filter_subsystems(udevadm_info, subsystems):
    """Filter udevadm_info, keep entries that match any of the passed subsystems."""

    ret = []
    for entry in udevadm_info:
        if entry["E"]["SUBSYSTEM"] in subsystems:
            ret.append(entry)
    return ret


def exportdb(subsystems=None):
    """Return the complete udev database.

    :param list subsystems: This parameter limits the returned data to specified
        subsystems such as "pci", "usb", "block", etc.

        .. versionadded :: 3008.0

    CLI Example:

    .. code-block:: bash

        salt '*' udev.exportdb
        salt '*' udev.exportdb subsystems='[usb, block]'

    """

    if subsystems is not None:
        if not isinstance(subsystems, list):
            raise SaltInvocationError("subsystems must be a list")

    cmd = "udevadm info --export-db"
    udev_result = __salt__["cmd.run_all"](cmd, output_loglevel="quiet")

    if udev_result["retcode"]:
        raise CommandExecutionError(udev_result["stderr"])

    ret = _parse_udevadm_info(udev_result["stdout"])
    if subsystems is not None:
        ret = _filter_subsystems(ret, subsystems)

    return ret



import pytest

import salt.modules.mac_pkgutil as mac_pkgutil
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {mac_pkgutil: {}}


def test_install():
    source = "/foo/bar/fubar.pkg"
    package_id = "com.foo.fubar.pkg"

    with patch("salt.modules.mac_pkgutil.is_installed", return_value=False):
        with patch(
            "salt.modules.mac_pkgutil._install_from_path", return_value=True
        ) as _install_from_path:
            mac_pkgutil.install(source, package_id)

    _install_from_path.assert_called_with(source)


def test_install_already_there():
    source = "/foo/bar/fubar.pkg"
    package_id = "com.foo.fubar.pkg"

    with patch("salt.modules.mac_pkgutil.is_installed", return_value=True):
        with patch(
            "salt.modules.mac_pkgutil._install_from_path", return_value=True
        ) as _install_from_path:
            mac_pkgutil.install(source, package_id)

    assert _install_from_path.called == 0



import pytest

import salt.modules.mac_keychain as keychain
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {keychain: {}}


def test_install_cert():
    """
    Test installing a certificate into the macOS keychain
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        keychain.install("/path/to/cert.p12", "passw0rd")
        mock.assert_called_once_with(
            "security import /path/to/cert.p12 -P passw0rd "
            "-k /Library/Keychains/System.keychain"
        )


def test_install_cert_extras():
    """
    Test installing a certificate into the macOS keychain with extras
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.mac_keychain.unlock_keychain"
    ) as unlock_mock:
        keychain.install(
            "/path/to/cert.p12",
            "passw0rd",
            "/path/to/chain",
            allow_any=True,
            keychain_password="passw0rd1",
        )
        unlock_mock.assert_called_once_with("/path/to/chain", "passw0rd1")
        mock.assert_called_once_with(
            "security import /path/to/cert.p12 -P passw0rd -k /path/to/chain -A"
        )


def test_uninstall_cert():
    """
    Test uninstalling a certificate from the macOS keychain
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        keychain.uninstall("/path/to/cert.p12", "passw0rd")
        mock.assert_called_once_with(
            'security delete-certificate -c "/path/to/cert.p12" passw0rd'
        )


def test_list_certs():
    """
    Test listing available certificates in a keychain
    """
    expected = ["com.apple.systemdefault", "com.apple.kerberos.kdc"]
    mock = MagicMock(return_value='"com.apple.systemdefault"\n"com.apple.kerberos.kdc"')
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        out = keychain.list_certs("/path/to/cert.p12")
        mock.assert_called_once_with(
            "security find-certificate -a /path/to/cert.p12 | "
            'grep -o "alis.*" | grep -o \'\\"[-A-Za-z0-9.:() ]*\\"\'',
            python_shell=True,
        )

        assert out == expected


def test_get_friendly_name():
    """
    Test getting the friendly name of a certificate
    """
    expected = "ID Installer Salt"
    mock = MagicMock(return_value="friendlyName: ID Installer Salt")
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        out = keychain.get_friendly_name("/path/to/cert.p12", "passw0rd", legacy=True)
        mock.assert_called_once_with(
            "openssl pkcs12 -legacy -in /path/to/cert.p12 -passin pass:passw0rd -info "
            "-nodes -nokeys 2> /dev/null | grep friendlyName:",
            python_shell=True,
        )

        assert out == expected

    mock = MagicMock(return_value="friendlyName: ID Installer Salt")
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        out = keychain.get_friendly_name("/path/to/cert.p12", "passw0rd", legacy=False)
        mock.assert_called_once_with(
            "openssl pkcs12 -in /path/to/cert.p12 -passin pass:passw0rd -info "
            "-nodes -nokeys 2> /dev/null | grep friendlyName:",
            python_shell=True,
        )

        assert out == expected


def test_get_default_keychain():
    """
    Test getting the default keychain
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        keychain.get_default_keychain("frank", "system")
        mock.assert_called_once_with(
            "security default-keychain -d system", runas="frank"
        )


def test_set_default_keychain():
    """
    Test setting the default keychain
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        keychain.set_default_keychain("/path/to/chain.keychain", "system", "frank")
        mock.assert_called_once_with(
            "security default-keychain -d system -s /path/to/chain.keychain",
            runas="frank",
        )


def test_unlock_keychain():
    """
    Test unlocking the keychain
    """
    mock = MagicMock()
    with patch.dict(keychain.__salt__, {"cmd.run": mock}):
        keychain.unlock_keychain("/path/to/chain.keychain", "passw0rd")
        mock.assert_called_once_with(
            "security unlock-keychain -p passw0rd /path/to/chain.keychain"
        )



"""
Tests for the archive state
"""

import errno
import logging
import os

import pytest

import salt.utils.files
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.helpers import Webserver
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS

# Setup logging
log = logging.getLogger(__name__)

ARCHIVE_DIR = (
    os.path.join("c:/", "tmp") if salt.utils.platform.is_windows() else "/tmp/archive"
)

ARCHIVE_NAME = "custom.tar.gz"
ARCHIVE_TAR_SOURCE = f"http://localhost:{9999}/{ARCHIVE_NAME}"
ARCHIVE_TAR_HASH = "md5=7643861ac07c30fe7d2310e9f25ca514"
ARCHIVE_TAR_SHA_HASH = (
    "sha256=9591159d86f0a180e4e0645b2320d0235e23e66c66797df61508bf185e0ac1d2"
)
ARCHIVE_TAR_BAD_HASH = "md5=d41d8cd98f00b204e9800998ecf8427e"
ARCHIVE_TAR_HASH_UPPER = "md5=7643861AC07C30FE7D2310E9F25CA514"


class ArchiveTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the archive state
    """

    @classmethod
    def setUpClass(cls):
        cls.webserver = Webserver()
        cls.webserver.start()
        cls.archive_tar_source = cls.webserver.url("custom.tar.gz")
        cls.archive_local_tar_source = "file://{}".format(
            os.path.join(RUNTIME_VARS.BASE_FILES, ARCHIVE_NAME)
        )
        cls.untar_file = os.path.join(ARCHIVE_DIR, "custom/README")

    @classmethod
    def tearDownClass(cls):
        cls.webserver.stop()

    def setUp(self):
        self._clear_archive_dir()

    def tearDown(self):
        self._clear_archive_dir()
        try:
            salt.utils.files.rm_rf(
                os.path.join(RUNTIME_VARS.TMP_ROOT_DIR, "cache", "archive_hash")
            )
        except OSError:
            # some tests do notcreate the archive_hash directory
            pass

    @staticmethod
    def _clear_archive_dir():
        try:
            salt.utils.files.rm_rf(ARCHIVE_DIR)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def _check_extracted(self, path):
        """
        function to check if file was extracted
        """
        log.debug("Checking for extracted file: %s", path)
        self.assertTrue(os.path.isfile(path))

    def run_function(self, *args, **kwargs):  # pylint: disable=arguments-differ
        ret = super().run_function(*args, **kwargs)
        log.debug("ret = %s", ret)
        return ret

    def run_state(self, *args, **kwargs):  # pylint: disable=arguments-differ
        ret = super().run_state(*args, **kwargs)
        log.debug("ret = %s", ret)
        return ret

    def test_archive_extracted_skip_verify(self):
        """
        test archive.extracted with skip_verify
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            archive_format="tar",
            skip_verify=True,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")
        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_with_source_hash(self):
        """
        test archive.extracted without skip_verify
        only external resources work to check to
        ensure source_hash is verified correctly
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_HASH,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.skip_if_not_root
    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_with_root_user_and_group(self):
        """
        test archive.extracted with user and group set to "root"
        """
        r_group = "root"
        if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
            r_group = "wheel"
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_HASH,
            user="root",
            group=r_group,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.slow_test
    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_with_strip_in_options(self):
        """
        test archive.extracted with --strip in options
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            source_hash=ARCHIVE_TAR_HASH,
            options="--strip=1",
            enforce_toplevel=False,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")

        self.assertSaltTrueReturn(ret)

        self._check_extracted(os.path.join(ARCHIVE_DIR, "README"))

    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_with_strip_components_in_options(self):
        """
        test archive.extracted with --strip-components in options
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            source_hash=ARCHIVE_TAR_HASH,
            options="--strip-components=1",
            enforce_toplevel=False,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")

        self.assertSaltTrueReturn(ret)

        self._check_extracted(os.path.join(ARCHIVE_DIR, "README"))

    @pytest.mark.slow_test
    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_without_archive_format(self):
        """
        test archive.extracted with no archive_format option
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            source_hash=ARCHIVE_TAR_HASH,
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")
        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.skip_on_fips_enabled_platform
    def test_archive_extracted_with_cmd_unzip_false(self):
        """
        test archive.extracted using use_cmd_unzip argument as false
        """

        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_tar_source,
            source_hash=ARCHIVE_TAR_HASH,
            use_cmd_unzip=False,
            archive_format="tar",
        )
        if "Timeout" in ret:
            self.skipTest("Timeout talking to local tornado server.")
        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    def test_local_archive_extracted(self):
        """
        test archive.extracted with local file
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.skip_on_fips_enabled_platform
    def test_local_archive_extracted_skip_verify(self):
        """
        test archive.extracted with local file, bad hash and skip_verify
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_BAD_HASH,
            skip_verify=True,
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.slow_test
    @pytest.mark.skip_on_fips_enabled_platform
    def test_local_archive_extracted_with_source_hash(self):
        """
        test archive.extracted with local file and valid hash
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_HASH,
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.slow_test
    @pytest.mark.skip_on_fips_enabled_platform
    def test_local_archive_extracted_with_bad_source_hash(self):
        """
        test archive.extracted with local file and bad hash
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_BAD_HASH,
        )

        self.assertSaltFalseReturn(ret)

    @pytest.mark.skip_on_fips_enabled_platform
    def test_local_archive_extracted_with_uppercase_source_hash(self):
        """
        test archive.extracted with local file and bad hash
        """
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            source_hash=ARCHIVE_TAR_HASH_UPPER,
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

    @pytest.mark.slow_test
    def test_archive_extracted_with_non_base_saltenv(self):
        """
        test archive.extracted with a saltenv other than `base`
        """
        ret = self.run_function(
            "state.sls",
            ["issue45893"],
            pillar={"issue45893.name": ARCHIVE_DIR},
            saltenv="prod",
        )
        self.assertSaltTrueReturn(ret)
        self._check_extracted(os.path.join(ARCHIVE_DIR, self.untar_file))

    @pytest.mark.slow_test
    def test_local_archive_extracted_with_skip_files_list_verify(self):
        """
        test archive.extracted with local file and skip_files_list_verify set to True
        """
        expected_comment = (
            "existing source sum is the same as the expected one and "
            "skip_files_list_verify argument was set to True. "
            "Extraction is not needed"
        )

        # Clearing the minion cache at the start to ensure that different tests of
        # skip_files_list_verify won't affect each other
        self.run_function("saltutil.clear_cache")
        self.run_function("saltutil.sync_all")

        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            skip_files_list_verify=True,
            source_hash_update=True,
            keep_source=True,
            source_hash=ARCHIVE_TAR_SHA_HASH,
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            skip_files_list_verify=True,
            source_hash_update=True,
            keep_source=True,
            source_hash=ARCHIVE_TAR_SHA_HASH,
        )

        self.assertSaltTrueReturn(ret)
        self.assertInSaltComment(expected_comment, ret)

    def test_local_archive_extracted_with_skip_files_list_verify_and_keep_source_is_false(
        self,
    ):
        """
        test archive.extracted with local file and skip_files_list_verify set to True
        and keep_source is set to False.
        """
        expected_comment = (
            "existing source sum is the same as the expected one and "
            "skip_files_list_verify argument was set to True. "
            "Extraction is not needed"
        )
        # Clearing the minion cache at the start to ensure that different tests of
        # skip_files_list_verify won't affect each other
        self.run_function("saltutil.clear_cache")
        self.run_function("saltutil.sync_all")

        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            skip_files_list_verify=True,
            source_hash_update=True,
            keep_source=False,
            source_hash=ARCHIVE_TAR_SHA_HASH,
        )

        self.assertSaltTrueReturn(ret)

        self._check_extracted(self.untar_file)

        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            skip_files_list_verify=True,
            source_hash_update=True,
            keep_source=False,
            source_hash=ARCHIVE_TAR_SHA_HASH,
        )

        self.assertSaltTrueReturn(ret)
        self.assertInSaltComment(expected_comment, ret)

    @pytest.mark.slow_test
    def test_local_archive_extracted_trim_output(self):
        """
        test archive.extracted with local file and trim_output set to 1
        """
        expected_changes = {
            "directories_created": ["/tmp/archive/"],
            "extracted_files": ["custom"],
        }
        ret = self.run_state(
            "archive.extracted",
            name=ARCHIVE_DIR,
            source=self.archive_local_tar_source,
            archive_format="tar",
            skip_files_list_verify=True,
            source_hash_update=True,
            source_hash=ARCHIVE_TAR_SHA_HASH,
            trim_output=1,
        )

        self.assertSaltTrueReturn(ret)
        self._check_extracted(self.untar_file)
        state_ret = ret["archive_|-/tmp/archive_|-/tmp/archive_|-extracted"]
        self.assertTrue(
            state_ret["comment"].endswith("Output was trimmed to 1 number of lines")
        )
        self.assertEqual(state_ret["changes"], expected_changes)
